/*
Copyright (C) 2004  The PARI group.

This file is part of the PARI package.

PARI/GP is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation. It is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY WHATSOEVER.

Check the License for details. You should have received a copy of it, along
with the package; see the file 'COPYING'. If not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

#include "pari.h"
#include "tree.h"
#include "stack.h"
stack s_node,s_value,s_func,s_ctx,s_aff,s_bloc,s_errors;

const char *opname[]={"_||_", "_&&_", "_==_", "_!=_", "_>=_", "_>_", "_<=_", "_<_", "_-_","_+_","_<<_", "_>>_", "_%_", "_\\/_", "_\\_", "_/_", "_*_","_^_","__","_--","_++","_-=_", "_+=_", "_<<=_", "_>>=_", "_%=_", "_\\/=_", "_\\=_", "_/=_", "_*=_","+_","-_","!_","_!","_'","_~","",""};


void initoperators(void)
{
  int i;
  for (i=0;i<OPnboperator;i++)
    newentry(opname[i]);
}
void
init_compiler(void)
{
  stack_init(&s_node,sizeof(*tree),(void **)&tree);
  stack_init(&s_value,sizeof(*value),(void **)&value);
  newnode(Fgnil,-1,-1);  /*GNIL*/
  newnode(Fgnil,-1,-1);  /*GNOARG*/
  initoperators();
}

GEN
eval(void)
{
  return evalnode (s_node.n-1);
}

int newnode(Ffunc f, int x, int y)
{
  int n=stack_new(&s_node);
  tree[n].f=f;
  tree[n].x=x;
  tree[n].y=y;
  tree[n].e=NULL;
  tree[n].args=NULL;
  tree[n].vars=NULL;
  return n;
}
int newopcall(OPerator op, int x, int y)
{
  if (y==-1)
    return newnode(Fentryfunc,op,x);
  else
    return newnode(Fentryfunc,op,newnode(Flistarg,x,y));
}

int newstringvalue(const char *s)
{
  int n=stack_new(&s_value);
  value[n].type=CSTstr;
  value[n].val.str=s;
  return n++;
}

int newquotevalue(const char *s)
{
  int n=stack_new(&s_value);
  value[n].type=CSTquote;
  value[n].val.str=s;
  return n++;
}

int newsmallrealvalue(long small)
{
  int n=stack_new(&s_value);
  value[n].type=CSTsmallreal;
  value[n].val.small=small;
  return n++;
}

int newgenvalue(GEN g)
{
  int n=stack_new(&s_value);
  value[n].type=CSTgen;
  value[n].val.gen=gclone(g);
  return n++;
}

int newdotvalue(GEN g)
{
  int n=stack_new(&s_value);
  value[n].type=CSTdot;
  value[n].val.gen=g;
  return n++;
}

extern int errno;

/* newxxxnode functions
 * These unctions take a token value as a string and a comment index.
 * They must return a new Fconst or Fsmall node.
 * They must free the string if they do not reference it.
 * the cast free((char *)s) is to remove the const qualifier.
 */
int newgennode(GEN g)
{
  if (typ(g)==t_INT)
  {
    if (signe(g)==0) return newnode(Fsmall,0,-1);
    if (lg(g)==3 && g[2]>=0)
      return newnode(Fsmall,g[2],-1);
  }
  return newnode(Fconst,newgenvalue(g),-1);
}

int newdotnode(GEN g)
{
  return newnode(Fconst,newdotvalue(g),-1);
}

int is_const(int n, CSTtype t)
{
  int f=tree[n].f;
  int x=tree[n].x;
  return f==Fconst && value[x].type==t;
}

int newentry(const char *s)
{
  return newstringvalue(s);
}

int isfunc(int n, const char *s)
{
  return tree[n].f==Fentryfunc && !strcmp(s,value[tree[n].x].val.str);
}

int newmember(const char *s)
{
  char *p;
  int n;
  p=calloc(sizeof(*s),3+strlen(s));
  sprintf(p,"_.%s",s);
  n=newentry(p);
  free((char *)s);
  return n;
}

int listtoseq(int *stack, int nb)
{
  int n,i;
  if (nb==-1) return -1;
  for(n=stack[0],i=1;i<nb;n=newnode(Flistarg,n,stack[i++]));
  return n;
}

void
printentryname(FILE *fout, const char *s)
{
  if (isdigit(*s))
    fprintf(fout,"p%s",s);
  else
    fprintf(fout,"%s",s);
}

void
printentry(FILE *fout, int x)
{
  const char *s=value[x].val.str;
  printentryname(fout, s);
}
void printnode(FILE *fout, int n);

int indent;
void printnodeparens(FILE *fout, int n, int parens)
{
  int x,y;
  if (n<0)
    return;
  x=tree[n].x;
  y=tree[n].y;
  switch(tree[n].f)
  {
  case Fseq:
    printnode(fout,x);
    if (x>=0 && tree[x].f==Fseq) x=tree[x].y;
    if (x>=0 && tree[x].f!=Fblock && tree[x].f!=Fdeffunc)
    {
      if (indent)
        fprintf(fout,";\n");
      else
        fprintf(fout,"; \\\n");
      //genindent(fout);
    }
    printnode(fout,y);
    break;
  case Fmatrix:
    fprintf(fout,"[");
    printnode(fout,x);
    fprintf(fout,"]");
    break;
  case FmatrixL:
    fprintf(fout,"[");
    printnode(fout,x);
    fprintf(fout,",]");
    break;
  case FmatrixR:
    fprintf(fout,"[,");
    printnode(fout,x);
    fprintf(fout,"]");
    break;
  case Ffacteurmat:
    printnodeparens(fout,x,1);
    printnode(fout,y);
    break;
  case Faffect:
    if (parens)
      fputc('(',fout);
    printnode(fout,x);
    fprintf(fout,"=");
    printnode(fout,y);
    if (parens)
      fputc(')',fout);
    break;
  case Fconst:
    switch(value[x].type)
    {
    case CSTsmall:
      fprintf(fout,"%ld",value[x].val.small);
      break;
    case CSTsmallreal:
      fprintf(fout,"%ld.",value[x].val.small);
      break;
    case CSTgen: 
      pariprintf("%Z",value[x].val.gen);
      break;
    case CSTdot: 
      pariprintf("%Z.",value[x].val.gen);
      break;
    case CSTstr:
      fprintf(fout,"\"%s\"",value[x].val.str);
      break;
    case CSTquote:
      fprintf(fout,"'%s",value[x].val.str);
      break;
    }
    break;
  case Fsmall:
    fprintf(fout,"%d",x);
    break;
  case Fmatrixelts:
    printnode(fout,x);
    fprintf(fout,",");
    printnode(fout,y);
    break;
  case Fmatrixlines:
    printnode(fout,x);
    fprintf(fout,";");
    /*Would be nice to indent matrix, alas I have no way to know where is the
     * bottomrow*/
    printnode(fout,y);
    break;
  case Femptyvec:
    fprintf(fout,"[]");
    break;
  case Femptymat:
    fprintf(fout,"[;]");
    break;
  case Fmat:
    fprintf(fout,"[");
    printnode(fout,x);
    fprintf(fout,"]");
    break;
  case Flistarg:
    printnode(fout,x);
    fprintf(fout,",");
    printnode(fout,y);
    break;
  case Fentry:
    printentry(fout,x);
    break;  
  case Fentryfunc:
    if (x<OPnboperator || (value[x].val.str[0]=='_' && value[x].val.str[1]=='.'))
    {
      int arg[1024];
      int nb=genlistargs(n,arg,0,1024);
      int i=0;
      const char *p, *name=value[x].val.str;
      if (parens && x!=OPcat)
	fputc('(',fout);
      for (p=name;*p;p++)
      {
	if (*p=='_')
	{
	  if (i==nb)
	    pari_err(talker,"too few arguments for operator %s",name);
	  printnodeparens(fout,arg[i++],1);

	}
	else
	  fputc(*p,fout);
      }
      if (i!=nb)
	 pari_err(talker,"too many arguments for operator %s",name);
      if (parens && x!=OPcat)
	fputc(')',fout);
    }
    else 
    {
      fprintf(fout,"%s",value[x].val.str);
      fprintf(fout,"(");
      if (y!=GNOARG)
        printnode(fout,y);
      fprintf(fout,")");
    }
    break;
  case Frefarg:
    fprintf(fout,"&%s",value[x].val.str);
    break;
  case Fdeffunc:
    fprintf(fout,"\n");
    printnode(fout,x);
    fprintf(fout,"=\n");
    //genindent(fout);
    if (tree[y].f!=Fblock)
    {
      fprintf(fout,"{\n");
      indent++;
      //genindent(fout);
    }
    printnode(fout,y);
    fprintf(fout,"\n");
    indent--;
    //genindent(fout);
    if (tree[y].f!=Fblock)
    {
      fprintf(fout,"}\n");
      //genindent(fout);
    }
    break;
  case Fgnil:
    if ( n!= GNOARG )
      fprintf(fout,"gnil");
    break;
  case Ftag:
    printnodeparens(fout,x,1);    
    break;
  }
}
void printnode(FILE *fout, int n)
{
  printnodeparens(fout,n,0);
}

const char *funcname(int f)
{
  if (f<0 || f>Flastfunc)
    return "Funknown";
  else
    return Ffuncname[f];
}
